/*
 * Copyright (c) 2013 Giam Teck Choon <choon@choon.net> or CHOON.NET.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <resolv.h>
#include <errno.h>
#include <libowfat/buffer.h>
#include <libowfat/io.h>
#include <libowfat/stralloc.h>
#include <libowfat/byte.h>
#include <libowfat/ip4.h>
#include <libowfat/ip6.h>
#include <libowfat/dns.h>

#include <msg/msg.h>
#include "mailfront.h"

#define eprintf(...) fprintf(stderr, __VA_ARGS__)
#define elogerr(N, M, ...) { if(N){ if(M){ eprintf_logname(); } fprintf(stderr, __VA_ARGS__); } }

#ifndef MAX_DNSTXT_RECORDS
#define MAX_DNSTXT_RECORDS 10
#endif
#ifndef DNS_MAX_DNSBL_LENGTH
#define DNS_MAX_DNSBL_LENGTH 512
#endif
#ifndef DNSBL_RWLDOMAINS_FILE
#define DNSBL_RWLDOMAINS_FILE "/var/qmail/control/rwldomains"
#endif
#ifndef DNSBL_RBLDOMAINS_FILE
#define DNSBL_RBLDOMAINS_FILE "/var/qmail/control/rbldomains"
#endif
#ifndef DNSBL_RWLDOMAINS_FILE_IPV6
#define DNSBL_RWLDOMAINS_FILE_IPV6 "/var/qmail/control/rwldomains6"
#endif
#ifndef DNSBL_RBLDOMAINS_FILE_IPV6
#define DNSBL_RBLDOMAINS_FILE_IPV6 "/var/qmail/control/rbldomains6"
#endif

static response resp;

static const char *mainprog = "mailfront";
static const char *thisprog = "plugin-check-dnsbl";
static unsigned long thispid;
static const char *linkproto;
static int is_ipv6 = 0;
static const char *local_ip;
static const char *remote_ip;
static const char *relayclient;
static const char *rbldomains;
static const char *rwldomains;
static const char *rbldomainsfile;
static const char *rwldomainsfile;
static const char *skipcheck;
static const char *rblsmtpd;
static const char *defercheck;
static const char *rejectcheck;
static const char *logcheck;
static str estr;
static str hmsg;

void eprintf_logname()
{
  if (!thispid) {
    thispid = (unsigned long)getpid();
  }
  eprintf("%s-%s[%lu]: ", mainprog, thisprog, thispid);
}

static inline char tohex(char c)
{
  return c>=10?c-10+'a':c+'0';
}

/* return 1 on success otherwise return 0 */
int reverse_ipstring(const char *ipstr, stralloc *sa)
{
  stralloc ip_reverse;
  stralloc_init(&ip_reverse);
  char hexval;
  int i, j;
  i = strlen(ipstr);
  if (byte_chr(ipstr, i, ':') < i) {
    char ip[16];
    if ((i=scan_ip6(ipstr, ip))==0) {
      return 0;
    }
    for (j=16; j>0; j--) {
      hexval=tohex(ip[j-1] & 15);
      if(!stralloc_catb(&ip_reverse, &hexval, 1)) return &resp_oom;
      if(!stralloc_cats(&ip_reverse, ".")) return &resp_oom;

      hexval=tohex(ip[j-1] >> 4);
      if(!stralloc_catb(&ip_reverse, &hexval, 1)) return &resp_oom;
      if(!stralloc_cats(&ip_reverse, ".")) return &resp_oom;
    }

  } else {
    while (i) {
      for (j = i;j > 0;--j) if (ipstr[j - 1] == '.') break;
      if (!stralloc_catb(&ip_reverse,ipstr + j,i - j)) return &resp_oom;
      if (!stralloc_cats(&ip_reverse,".")) return &resp_oom;
      if (!j) break;
      i = j - 1;
    }
  }
  stralloc_0(&ip_reverse);
  if (!stralloc_copyb(sa, ip_reverse.s, ip_reverse.len)) return &resp_oom;
  stralloc_free(&ip_reverse);
  return 1;
}

/* return 0 on success otherwise return errno=ENOMEM
 * read file content and copy to out without newline
 */
int ioread_file_append_sa(const char *filename, stralloc *out)
{
  int64 fd;
  if (!io_readfile(&fd, filename)) return 0;
  static stralloc line;
  buffer in;
  char buf[64];
  long r;
  buffer_init(&in,read,fd,buf,sizeof buf);
  while ((r=buffer_getnewline_sa(&in, &line))>0) {
    if (line.len > 1) {
      if (line.s[0] != '#' && line.s[0] != ' ') {
        if (out->len > 0) {
          /* append a space */
          if (!stralloc_cats(out, " ")) {
            errno = ENOMEM;
            return errno;
          }
        }
        if (!stralloc_catb(out, line.s, line.len-1)) {
          errno = ENOMEM;
          return errno;
        }
      }
    }
  }
  return 0;
}

struct HostDNSTXT {
  char *txtname;
  int size;
  char *txtrecord[MAX_DNSTXT_RECORDS];
};

void init_HostDNSTXT(struct HostDNSTXT *d)
{
  d->txtname = NULL;
  d->size = 0;
}

void free_HostDNSTXT(struct HostDNSTXT *d)
{
  int x = 0;
  if (d->size > 0) {
    for (x = 0; x < d->size; x++) {
      free(d->txtrecord[x]);
    }
    d->size = 0;
  }
  init_HostDNSTXT(d);
}

/* return 0 for success otherwise return errno=ENOMEM if out of memory */
int check_dns_txt(const char *name, struct HostDNSTXT *s, int limit)
{
  unsigned char *ansbuf = (unsigned char *)malloc(NS_PACKETSZ);
  if (ansbuf == NULL) {
    errno = ENOMEM;
    return errno;
  }

  int l, i, c;
  c = 0;

  l = res_query(name, C_IN, T_TXT, ansbuf, NS_PACKETSZ);
  if (l > 0 && ((HEADER *)ansbuf)->ancount != 0) {
    unsigned char *recbuf = ansbuf+NS_HFIXEDSZ;
    for (i = ns_get16(ansbuf+4); i != 0; --i) {
      recbuf += dn_skipname(recbuf, ansbuf+l)+4;
    }
    for (i = ns_get16(ansbuf+6); i != 0; --i) {

      recbuf += dn_skipname(recbuf, ansbuf+l);

      if (ns_get16(recbuf) != T_TXT) {
        recbuf += 10 + ns_get16(recbuf+8);
        continue;
      }
      /* it's a TXT record, wow */
      if (c < limit) {
        if (s->txtname == NULL) {
          s->txtname = (char *)name;
        }
        char *tmp = (char *)malloc(256);
        if (tmp == NULL) {
          errno = ENOMEM;
          free(ansbuf);
          return errno;
        }
        snprintf(tmp, recbuf[10]+1, "%s", (char *)recbuf+11);
        s->txtrecord[c] = strdup(tmp);
        if (s->txtrecord[c] == NULL) {
          errno = ENOMEM;
          free(ansbuf);
          free(tmp);
          return errno;
        }
        free(tmp);
        c++;
      }
    }
  }
  free(ansbuf);
  s->size = c;
  return 0;
}

/* return 0 on success otherwise return errno=ENOMEM */
int read_dnsbwlfiles()
{
  if (!rwldomains && rwldomainsfile && *rwldomainsfile) {
    stralloc rwl;
    stralloc_init(&rwl);
    if (!stralloc_copys(&rwl, "")) {
      errno = ENOMEM;
      elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
      return errno;
    }
    int r = ioread_file_append_sa(rwldomainsfile, &rwl);
    if (r && errno) return errno;
    if (rwl.len > 0 && strlen(rwl.s) > 0) {
      stralloc_0(&rwl);
      rwldomains = strdup(rwl.s);
      if (rwldomains == NULL) {
        errno = ENOMEM;
        elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
        stralloc_free(&rwl);
        return errno;
      }
    }
  }
  if (!rbldomains && rbldomainsfile && *rbldomainsfile) {
    stralloc rbl;
    stralloc_init(&rbl);
    if (!stralloc_copys(&rbl, "")) {
      errno = ENOMEM;
      elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
      return errno;
    }
    int r = ioread_file_append_sa(rbldomainsfile, &rbl);
    if (r && errno) return errno;
    if (rbl.len > 0 && strlen(rbl.s) > 0) {
      stralloc_0(&rbl);
      rbldomains = strdup(rbl.s);
      if (rbldomains == NULL) {
        errno = ENOMEM;
        elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
        stralloc_free(&rbl);
        return errno;
      }
    }
  }
  return 0;
}

/* If $RBLSMTPD is set and not empty, first char is - then reply with 553
 * otherwise reply with 451
 * return 1 to block otherwise return 0
 */
int check_rblsmtpd()
{
  const char *x = rblsmtpd;
  if (x) {
    if (!*x) return 0;
    if (*x == '-') {
      resp.number = 553;
      ++x;
    } else {
      resp.number = 451;
    }
    resp.message = x;
    return 1;
  }
  return 0;
}

static const response *init(void)
{
  linkproto = getprotoenv(0);
  is_ipv6 = linkproto != 0 && strcasecmp(linkproto, "TCP6") == 0;
  local_ip = getprotoenv("LOCALIP");
  remote_ip = getprotoenv("REMOTEIP");
  if (remote_ip && *remote_ip) {
    if (byte_chr(remote_ip, strlen(remote_ip), ':') < strlen(remote_ip))
      is_ipv6 = 1;
  }
  relayclient = session_getenv("RELAYCLIENT");
  rejectcheck = session_getenv("DNSBL_CHECK_REJECT");
  defercheck = session_getenv("DNSBL_CHECK_DEFER");
  skipcheck = session_getenv("DNSBL_CHECK_SKIP");
  rblsmtpd = session_getenv("RBLSMTPD");
  logcheck = session_getenv("DNSBL_CHECK_VERBOSE");
  rwldomains = session_getenv("RWLDOMAINS");
  rbldomains = session_getenv("RBLDOMAINS");
  rwldomainsfile = session_getenv("RWLDOMAINSFILE");
  if (!rwldomainsfile) {
    if (is_ipv6)
      rwldomainsfile = DNSBL_RWLDOMAINS_FILE_IPV6;
    else
      rwldomainsfile = DNSBL_RWLDOMAINS_FILE;
  }
  rbldomainsfile = session_getenv("RBLDOMAINSFILE");
  if (!rbldomainsfile) {
    if (is_ipv6)
      rbldomainsfile = DNSBL_RBLDOMAINS_FILE_IPV6;
    else
      rbldomainsfile = DNSBL_RBLDOMAINS_FILE;
  }
  thispid = (unsigned long)getpid();
  if (read_dnsbwlfiles() != 0) {
    return &resp_oom;
  }
  return 0;
}

static const response *sender(str* address, str* params)
{
  if (skipcheck) return 0;
  if (!local_ip) {
    local_ip = getprotoenv("LOCALIP");
  }
  if (!local_ip) {
    elogerr(logcheck, 1, "LOCALIP not found thus this plugin skipped\n");
    return 0;
  }
  if (!remote_ip) {
    remote_ip = getprotoenv("REMOTEIP");
  }
  if (!remote_ip) {
    elogerr(logcheck, 1, "REMOTEIP not found thus this plugin skipped\n");
    return 0;
  }
  str_init(&hmsg);
  /* X-DNSBL-Check-Result: SKIPPED: $RELAYCLIENT is set for 1.2.3.4 */
  if (relayclient) {
    elogerr(logcheck, 1, "$RELAYCLIENT is set thus skipped for REMOTEIP %s\n", remote_ip);
    str_cats(&hmsg, "X-DNSBL-Check-Result: SKIPPED: $RELAYCLIENT is set for ");
    str_cats(&hmsg, remote_ip);
    str_cats(&hmsg, "\n");
    return 0;
  }
  /* X-DNSBL-Check-Result: SKIPPED: authenticated user from 1.2.3.4 */
  if (session_getnum("authenticated", 0)) {
    elogerr(logcheck, 1, "authenticated user thus skipped for REMOTEIP %s\n", remote_ip);
    str_cats(&hmsg, "X-DNSBL-Check-Result: SKIPPED: authenticated user from ip ");
    str_cats(&hmsg, remote_ip);
    str_cats(&hmsg, "\n");
    return 0;
  }
  /* X-DNSBL-Check-Result: SKIPPED: $RBLSMTPD is set and empty for 1.1.1.1 */
  if (rblsmtpd && rblsmtpd[0] == 0) {
    /* if is set and empty we skip it */
    elogerr(logcheck, 1, "$RBLSMTPD is set and empty thus skipped for REMOTEIP %s\n", remote_ip);
    str_cats(&hmsg, "X-DNSBL-Check-Result: SKIPPED: $RBLSMTPD is set and empty for ");
    str_cats(&hmsg, remote_ip);
    str_cats(&hmsg, "\n");
    return 0;
  }
  if (check_rblsmtpd()) {
    return &resp;
  }
  /* X-DNSBL-Check-Result: SKIPPED: $RWLDOMAINS and $RBLDOMAINS not set or empty for 1.2.3.4 */
  if ((!rwldomains && !rbldomains) || (!*rwldomains && !*rbldomains)) {
    /* if is set and empty we skip it */
    elogerr(logcheck, 1, "$RWLDOMAINS and $RBLDOMAINS not set or empty thus skipped for REMOTEIP %s\n", remote_ip);
    str_cats(&hmsg, "X-DNSBL-Check-Result: SKIPPED: $RWLDOMAINS and $RBLDOMAINS not set or empty for ");
    str_cats(&hmsg, remote_ip);
    str_cats(&hmsg, "\n");
    return 0;
  }
  int action = 0;
  if (rejectcheck || defercheck) {
    if (rejectcheck) {
      action = 553;
      resp.number = action;
    } else {
      action = 451;
      resp.number = action;
    }
  }
  char *delimiters = " ,\t\n";
  stralloc ip_reverse;
  stralloc_init(&ip_reverse);
  if (!reverse_ipstring(remote_ip, &ip_reverse) || !ip_reverse.len) {
    elogerr(1, 1, "(%s:%d) problem generating ip reverse format for DNSBL lookup\n", __FILE__, __LINE__);
    return 0;
    //return &resp_internal;
  }
  int rwl_listed = 0;
  int rbl_listed = 0;
  str_init(&estr);
  stralloc tmprbl;
  stralloc_zero(&tmprbl);
  stralloc_init(&tmprbl);
  /* rwldomains */
  if (rwldomains && *rwldomains) {
    char *pch;
    char *tmppch = strdup(rwldomains);
    if (tmppch == NULL) {
      errno = ENOMEM;
      elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
      return &resp_oom;
    }
    pch = strtok(tmppch, delimiters);
    while (pch != NULL) {
      if (*pch != '#' && strlen(pch) > 1) {
        stralloc_zero(&tmprbl);
        stralloc_init(&tmprbl);
        if (!stralloc_copys(&tmprbl, ip_reverse.s)) {
          errno = ENOMEM;
          elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
          free(tmppch);
          return &resp_oom;
        }
        if (!stralloc_cats(&tmprbl, pch)) {
          errno = ENOMEM;
          elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
          free(tmppch);
          return &resp_oom;
        }
        stralloc_0(&tmprbl);
        elogerr(logcheck, 1, "Checking rwldomain=%s listing for ip=%s query=%s\n", pch, remote_ip, tmprbl.s);
        struct HostDNSTXT txtres;
        init_HostDNSTXT(&txtres);
        int rc = check_dns_txt(tmprbl.s, &txtres, MAX_DNSTXT_RECORDS);
        int i = 0;
        if (rc != 0 || errno == ENOMEM) {
          elogerr(1, 1, "(%s:%d) check_dns_txt Out of Memory!\n", __FILE__, __LINE__);
          free(tmppch);
          free_HostDNSTXT(&txtres);
          return &resp_oom;
        }
        for (i = 0; i < txtres.size; i++) {
          if (txtres.txtrecord[i]) {
            elogerr(logcheck, 1, "rwldomain=%s ip=%s query=%s result=%s\n", pch, remote_ip, tmprbl.s, txtres.txtrecord[i]);
            str_cats(&estr, "X-DNSWL-Check-Report: ");
            str_cats(&estr, "rwldomain=");
            str_cats(&estr, pch);
            str_cats(&estr, " result=");
            str_cats(&estr, txtres.txtrecord[i]);
            str_cats(&estr, "\n");
            rwl_listed++;
          }
        }
        free_HostDNSTXT(&txtres);
      }
      pch = strtok(NULL, delimiters);
    }
    free(tmppch);
  }
  /* rbldomains */
  stralloc rblmsg;
  stralloc_init(&rblmsg);
  if (rbldomains && *rbldomains) {
    char *pch;
    char *tmppch = strdup(rbldomains);
    if (tmppch == NULL) {
      errno = ENOMEM;
      elogerr(logcheck, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
      return &resp_oom;
    }
    pch = strtok(tmppch, delimiters);
    while (pch != NULL) {
      if (*pch != '#' && strlen(pch) > 1) {
        stralloc_zero(&tmprbl);
        stralloc_init(&tmprbl);
        if (!stralloc_copys(&tmprbl, ip_reverse.s)) {
          errno = ENOMEM;
          elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
          free(tmppch);
          return &resp_oom;
        }
        if (!stralloc_cats(&tmprbl, pch)) {
          errno = ENOMEM;
          elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
          free(tmppch);
          return &resp_oom;
        }
        stralloc_0(&tmprbl);
        elogerr(logcheck, 1, "Checking rbldomain=%s listing for ip=%s query=%s\n", pch, remote_ip, tmprbl.s);
        struct HostDNSTXT txtres;
        init_HostDNSTXT(&txtres);
        int rc = check_dns_txt(tmprbl.s, &txtres, MAX_DNSTXT_RECORDS);
        int i = 0;
        if (rc != 0 || errno == ENOMEM) {
          elogerr(logcheck, 1, "(%s:%d) check_dns_txt Out of Memory!\n", __FILE__, __LINE__);
          free(tmppch);
          free_HostDNSTXT(&txtres);
          return &resp_oom;
        }
        for (i = 0; i < txtres.size; i++) {
          if (txtres.txtrecord[i]) {
            elogerr(logcheck, 1, "rbldomain=%s ip=%s query=%s result=%s\n", pch, remote_ip, tmprbl.s, txtres.txtrecord[i]);
            str_cats(&estr, "X-DNSBL-Check-Report: ");
            str_cats(&estr, "rbldomain=");
            str_cats(&estr, pch);
            str_cats(&estr, " result=");
            str_cats(&estr, txtres.txtrecord[i]);
            str_cats(&estr, "\n");
            if (rbl_listed == 0) {
              if (!stralloc_cats(&rblmsg, "Sorry, your ip ")) {
                elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
                return &resp_oom;
              }
              if (!stralloc_cats(&rblmsg, remote_ip)) {
                elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
                return &resp_oom;
              }
              if (!stralloc_cats(&rblmsg, " is listed in DNSBL ")) {
                elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
                return &resp_oom;
              }
              if (!stralloc_cats(&rblmsg, pch)) {
                elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
                return &resp_oom;
              }
              if (!stralloc_cats(&rblmsg, " and its txt record is ")) {
                elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
                return &resp_oom;
              }
              if (!stralloc_cats(&rblmsg, txtres.txtrecord[i])) {
                elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
                return &resp_oom;
              }
              stralloc_0(&rblmsg);
            }
            rbl_listed++;
          }
        }
        free_HostDNSTXT(&txtres);
      }
      pch = strtok(NULL, delimiters);
    }
    free(tmppch);
  }
  str_init(&hmsg);
  str_cats(&hmsg, "X-DNSBL-Check-Result: ");
  if (rwl_listed > 0) {
    str_cats(&hmsg, "PASSED for ");
    action = 0;
  } else {
    if (rbl_listed > 0) {
      str_cats(&hmsg, "FAILED for ");
    } else {
      str_cats(&hmsg, "PASSED for ");
      action = 0;
    }
  }
  str_cats(&hmsg, remote_ip);
  str_cats(&hmsg, "\n");
  if (estr.len > 0 && estr.s) {
    str_cats(&hmsg, estr.s);
    str_init(&estr);
  }
  if (action > 0 && rbldomains && *rbldomains) {
    if (!rblmsg.len) {
      if (!stralloc_cats(&rblmsg, "Sorry, your ip ")) {
        elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
        return &resp_oom;
      }
      if (!stralloc_cats(&rblmsg, remote_ip)) {
        elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
        return &resp_oom;
      }
      if (!stralloc_cats(&rblmsg, " is listed in one or more of the DNSBL: ")) {
        elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
        return &resp_oom;
      }
      if (!stralloc_cats(&rblmsg, rbldomains)) {
        elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
        return &resp_oom;
      }
      stralloc_0(&rblmsg);
    }
    str_init(&estr);
    str_cats(&estr, rblmsg.s);
    resp.message = estr.s;
    stralloc_free(&tmprbl);
    str_init(&hmsg);
    return &resp;
  }
  stralloc_free(&tmprbl);
  return 0;
  (void)address;
  (void)params;
}

static const response *data_start(int fd)
{
  if (skipcheck) return 0;
  if (hmsg.len > 0) {
    return backend_data_block(hmsg.s, hmsg.len);
  }
  return 0;
  (void)fd;
}

struct plugin plugin = {
  .version = PLUGIN_VERSION,
  .flags = 0,
  .init = init,
  .sender = sender,
  .data_start = data_start,
};
